@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2000, 2004 IBM Corporation and others.
+ * Copyright (c) 2000, 2006 IBM Corporation and others.
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the Eclipse Public License v1.0
  * which accompanies this distribution, and is available at
@@ -13,15 +13,33 @@ package org.eclipse.jdt.internal.compiler.env;
 import org.eclipse.jdt.internal.compiler.impl.Constant;
 
 public interface IBinaryField extends IGenericField {
+/**
+ * Answer the runtime visible and invisible annotations for this field or null if none.
+ */
+IBinaryAnnotation[] getAnnotations();
+
 /**
  * 
  * @return org.eclipse.jdt.internal.compiler.Constant
  */
 Constant getConstant();
+
+/**
+ * Answer the receiver's signature which describes the parameter &
+ * return types as specified in section 4.4.4 of the Java 2 VM spec.
+ */
+char[] getGenericSignature();
+
 /**
  * Answer the name of the field.
  */
 char[] getName();
+
+/**
+ * Answer the tagbits set according to the bits for annotations.
+ */
+long getTagBits();
+
 /**
  * Answer the resolved name of the receiver's type in the
  * class file format as specified in section 4.3.2 of the Java 2 VM spec.
@@ -32,16 +50,5 @@ char[] getName();
  *   - a 2 dimensional array of strings is [[Ljava/lang/String;
  *   - an array of floats is [F
  */
-
 char[] getTypeName();
-
-/**
- * Answer the receiver's signature which describes the parameter &
- * return types as specified in section 4.4.4 of the Java 2 VM spec.
- */
-char[] getGenericSignature();
-/**
- * Answer the tagbits set according to the bits for annotations.
- */
-long getTagBits();
 }
