@@ -1,5 +1,5 @@
 /*******************************************************************************
- * Copyright (c) 2000, 2005 IBM Corporation and others.
+ * Copyright (c) 2000, 2006 IBM Corporation and others.
  * All rights reserved. This program and the accompanying materials
  * are made available under the terms of the Eclipse Public License v1.0
  * which accompanies this distribution, and is available at
@@ -12,17 +12,20 @@ package org.eclipse.jdt.internal.compiler.lookup;
 
 import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
 import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
+import org.eclipse.jdt.internal.compiler.classfmt.ClassFileConstants;
 import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
 import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
 import org.eclipse.jdt.internal.compiler.util.HashtableOfObject;
+import org.eclipse.jdt.internal.compiler.util.SimpleSet;
 
-public class MethodVerifier implements TagBits, TypeConstants {
+public class MethodVerifier {
 	SourceTypeBinding type;
 	HashtableOfObject inheritedMethods;
 	HashtableOfObject currentMethods;
 	ReferenceBinding runtimeException;
 	ReferenceBinding errorException;
 	LookupEnvironment environment;
+	private boolean allowCompatibleReturnTypes;
 /*
 Binding creation is responsible for reporting all problems with types:
 	- all modifier problems (duplicates & multiple visibility modifiers + incompatible combinations - abstract/final)
@@ -47,6 +50,9 @@ MethodVerifier(LookupEnvironment environment) {
 	this.runtimeException = null;
 	this.errorException = null;
 	this.environment = environment;
+	this.allowCompatibleReturnTypes =
+		environment.globalOptions.complianceLevel >= ClassFileConstants.JDK1_5
+			&& environment.globalOptions.sourceLevel < ClassFileConstants.JDK1_5;
 }
 boolean areMethodsEqual(MethodBinding one, MethodBinding two) {
 	return doesMethodOverride(one, two) && areReturnTypesEqual(one, two);
@@ -63,8 +69,41 @@ boolean areParametersEqual(MethodBinding one, MethodBinding two) {
 		if (!areTypesEqual(oneArgs[i], twoArgs[i])) return false;
 	return true;
 }
+boolean areReturnTypesCompatible(MethodBinding one, MethodBinding substituteTwo) {
+	if (one.returnType == substituteTwo.returnType) return true;
+
+	// short is compatible with int, but as far as covariance is concerned, its not
+	if (one.returnType.isBaseType()) return false;
+
+	if (!one.declaringClass.isInterface()) {
+		if (one.declaringClass.id == TypeIds.T_JavaLangObject)
+			return substituteTwo.returnType.isCompatibleWith(one.returnType); // interface methods inherit from Object
+		return one.returnType.isCompatibleWith(substituteTwo.returnType);
+	}
+
+	// check for methods from Object, every interface inherits from Object
+	if (substituteTwo.declaringClass.id == TypeIds.T_JavaLangObject)
+		return one.returnType.isCompatibleWith(substituteTwo.returnType);
+
+	// both are interfaces, see if they're related
+	if (one.declaringClass.implementsInterface(substituteTwo.declaringClass, true))
+		return one.returnType.isCompatibleWith(substituteTwo.returnType);
+	if (substituteTwo.declaringClass.implementsInterface(one.declaringClass, true))
+		return substituteTwo.returnType.isCompatibleWith(one.returnType);
+
+	// unrelated interfaces... one must be a subtype of the other
+	return one.returnType.isCompatibleWith(substituteTwo.returnType)
+		|| substituteTwo.returnType.isCompatibleWith(one.returnType);
+}
 boolean areReturnTypesEqual(MethodBinding one, MethodBinding two) {
-	return areTypesEqual(one.returnType, two.returnType);
+	if (areTypesEqual(one.returnType, two.returnType)) return true;
+
+	// when sourceLevel < 1.5 but compliance >= 1.5, allow return types in binaries to be compatible instead of just equal
+	if (this.allowCompatibleReturnTypes)
+		if (one.declaringClass instanceof BinaryTypeBinding && two.declaringClass instanceof BinaryTypeBinding)
+			if (areReturnTypesCompatible(one, two))
+				return true;
+	return false;
 }
 boolean areTypesEqual(TypeBinding one, TypeBinding two) {
 	if (one == two) return true;
@@ -82,7 +121,7 @@ boolean areTypesEqual(TypeBinding one, TypeBinding two) {
 boolean canSkipInheritedMethods() {
 	if (this.type.superclass() != null && this.type.superclass().isAbstract())
 		return false;
-	return this.type.superInterfaces() == NoSuperInterfaces;
+	return this.type.superInterfaces() == Binding.NO_SUPERINTERFACES;
 }
 boolean canSkipInheritedMethods(MethodBinding one, MethodBinding two) {
 	return two == null // already know one is not null
@@ -115,24 +154,23 @@ void checkAgainstInheritedMethods(MethodBinding currentMethod, MethodBinding[] m
 		// want to tag currentMethod even if return types are not equal
 		if (inheritedMethod.isAbstract()) {
 			if (inheritedMethod.declaringClass.isInterface()) {
-				currentMethod.modifiers |= CompilerModifiers.AccImplementing;
+				currentMethod.modifiers |= ExtraCompilerModifiers.AccImplementing;
 			} else {
-				currentMethod.modifiers |= CompilerModifiers.AccImplementing | CompilerModifiers.AccOverriding;
+				currentMethod.modifiers |= ExtraCompilerModifiers.AccImplementing | ExtraCompilerModifiers.AccOverriding;
 			}
 // with the above change an abstract method is tagged as implementing the inherited abstract method
 //		if (!currentMethod.isAbstract() && inheritedMethod.isAbstract()) {
 //			if ((currentMethod.modifiers & CompilerModifiers.AccOverriding) == 0)
 //				currentMethod.modifiers |= CompilerModifiers.AccImplementing;
 		} else {
-			currentMethod.modifiers |= CompilerModifiers.AccOverriding;
+			currentMethod.modifiers |= ExtraCompilerModifiers.AccOverriding;
 		}
 
-		if (!areReturnTypesEqual(currentMethod, inheritedMethod)) {
-			problemReporter(currentMethod).incompatibleReturnType(currentMethod, inheritedMethod);
-			continue nextMethod;
-		}
+		if (!areReturnTypesEqual(currentMethod, inheritedMethod))
+			if (reportIncompatibleReturnTypeError(currentMethod, inheritedMethod))
+				continue nextMethod;
 
-		if (currentMethod.thrownExceptions != NoExceptions)
+		if (currentMethod.thrownExceptions != Binding.NO_EXCEPTIONS)
 			checkExceptions(currentMethod, inheritedMethod);
 		if (inheritedMethod.isFinal())
 			problemReporter(currentMethod).finalMethodCannotBeOverridden(currentMethod, inheritedMethod);
@@ -161,7 +199,7 @@ void checkConcreteInheritedMethod(MethodBinding concreteMethod, MethodBinding[]
 	if (!concreteMethod.isPublic())
 		// Cannot reduce visibility of a public method specified by an interface
 		problemReporter().inheritedMethodReducesVisibility(type, concreteMethod, abstractMethods);
-	if (concreteMethod.thrownExceptions != NoExceptions)
+	if (concreteMethod.thrownExceptions != Binding.NO_EXCEPTIONS)
 		for (int i = abstractMethods.length; --i >= 0;)
 			checkExceptions(concreteMethod, abstractMethods[i]);
 }
@@ -187,17 +225,8 @@ void checkForBridgeMethod(MethodBinding currentMethod, MethodBinding inheritedMe
 	// no op before 1.5
 }
 void checkInheritedMethods(MethodBinding[] methods, int length) {
-	MethodBinding first = methods[0];
-	int index = length;
-	while (--index > 0 && areReturnTypesEqual(first, methods[index])){/*empty*/}
-	if (index > 0) {  // All inherited methods do NOT have the same vmSignature
-		if (this.type.isInterface())
-			for (int i = length; --i >= 0;)
-				if (methods[i].declaringClass.id == TypeIds.T_JavaLangObject)
-					return; // do not complain since the super interface already got blamed
-		problemReporter().inheritedMethodsHaveIncompatibleReturnTypes(this.type, methods, length);
+	if (!checkInheritedReturnTypes(methods, length))
 		return;
-	}
 
 	MethodBinding concreteMethod = null;
 	if (!type.isInterface()) {  // ignore concrete methods for interfaces
@@ -227,12 +256,27 @@ void checkInheritedMethods(MethodBinding[] methods, int length) {
 	}
 
 	MethodBinding[] abstractMethods = new MethodBinding[length - 1];
-	index = 0;
+	int index = 0;
 	for (int i = length; --i >= 0;)
 		if (methods[i] != concreteMethod)
 			abstractMethods[index++] = methods[i];
 	checkConcreteInheritedMethod(concreteMethod, abstractMethods);
 }
+boolean checkInheritedReturnTypes(MethodBinding[] methods, int length) {
+	MethodBinding first = methods[0];
+	int index = length;
+	while (--index > 0 && areReturnTypesEqual(first, methods[index])){/*empty*/}
+	if (index == 0) 
+		return true;
+
+	// All inherited methods do NOT have the same vmSignature
+	if (this.type.isInterface())
+		for (int i = length; --i >= 0;)
+			if (methods[i].declaringClass.id == TypeIds.T_JavaLangObject)
+				return false; // do not complain since the super interface already got blamed
+	problemReporter().inheritedMethodsHaveIncompatibleReturnTypes(this.type, methods, length);
+	return false;
+}
 /*
 For each inherited method identifier (message pattern - vm signature minus the return type)
 	if current method exists
@@ -365,11 +409,13 @@ void computeInheritedMethods(ReferenceBinding superclass, ReferenceBinding[] sup
 	// if an inheritedMethod has been 'replaced' by a supertype's method then skip it
 
 	this.inheritedMethods = new HashtableOfObject(51); // maps method selectors to an array of methods... must search to match paramaters & return type
-	ReferenceBinding[][] interfacesToVisit = new ReferenceBinding[3][];
-	int lastPosition = -1;
+	ReferenceBinding[] interfacesToVisit = null;
+	int nextPosition = 0;
 	ReferenceBinding[] itsInterfaces = superInterfaces;
-	if (itsInterfaces != NoSuperInterfaces)
-		interfacesToVisit[++lastPosition] = itsInterfaces;
+	if (itsInterfaces != Binding.NO_SUPERINTERFACES) {
+		nextPosition = itsInterfaces.length;
+		interfacesToVisit = itsInterfaces;
+	}
 
 	ReferenceBinding superType = superclass;
 	HashtableOfObject nonVisibleDefaultMethods = new HashtableOfObject(3); // maps method selectors to an array of methods
@@ -379,10 +425,21 @@ void computeInheritedMethods(ReferenceBinding superclass, ReferenceBinding[] sup
 	    if (allSuperclassesAreAbstract) {
 		    if (superType.isAbstract()) {
 				// only need to include superinterfaces if immediate superclasses are abstract
-				if ((itsInterfaces = superType.superInterfaces()) != NoSuperInterfaces) {
-					if (++lastPosition == interfacesToVisit.length)
-						System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[lastPosition * 2][], 0, lastPosition);
-					interfacesToVisit[lastPosition] = itsInterfaces;
+				if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES) {
+					if (interfacesToVisit == null) {
+						interfacesToVisit = itsInterfaces;
+						nextPosition = interfacesToVisit.length;
+					} else {
+						int itsLength = itsInterfaces.length;
+						if (nextPosition + itsLength >= interfacesToVisit.length)
+							System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
+						nextInterface : for (int a = 0; a < itsLength; a++) {
+							ReferenceBinding next = itsInterfaces[a];
+							for (int b = 0; b < nextPosition; b++)
+								if (next == interfacesToVisit[b]) continue nextInterface;
+							interfacesToVisit[nextPosition++] = next;
+						}
+					}
 				}
 			} else {
 			    allSuperclassesAreAbstract = false;
@@ -445,47 +502,43 @@ void computeInheritedMethods(ReferenceBinding superclass, ReferenceBinding[] sup
 		}
 		superType = superType.superclass();
 	}
+	if (nextPosition == 0) return;
 
-	for (int i = 0; i <= lastPosition; i++) {
-		ReferenceBinding[] interfaces = interfacesToVisit[i];
-		for (int j = 0, l = interfaces.length; j < l; j++) {
-			superType = interfaces[j];
-			if ((superType.tagBits & InterfaceVisited) == 0) {
-				superType.tagBits |= InterfaceVisited;
-				if (superType.isValidBinding()) {
-					if ((itsInterfaces = superType.superInterfaces()) != NoSuperInterfaces) {
-						if (++lastPosition == interfacesToVisit.length)
-							System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[lastPosition * 2][], 0, lastPosition);
-						interfacesToVisit[lastPosition] = itsInterfaces;
-					}
-
-					MethodBinding[] methods = superType.unResolvedMethods();
-					nextMethod : for (int m = methods.length; --m >= 0;) { // Interface methods are all abstract public
-						MethodBinding inheritedMethod = methods[m];
-						MethodBinding[] existingMethods = (MethodBinding[]) this.inheritedMethods.get(inheritedMethod.selector);
-						if (existingMethods == null) {
-							existingMethods = new MethodBinding[] {inheritedMethod};
-						} else {
-							int length = existingMethods.length;
-							// look to see if any of the existingMethods implement this inheritedMethod
-							for (int e = 0; e < length; e++)
-								if (isInterfaceMethodImplemented(inheritedMethod, existingMethods[e], superType))
-									continue nextMethod; // skip interface method with the same signature if visible to its declaringClass
-							System.arraycopy(existingMethods, 0, existingMethods = new MethodBinding[length + 1], 0, length);
-							existingMethods[length] = inheritedMethod;
-						}
-						this.inheritedMethods.put(inheritedMethod.selector, existingMethods);
-					}
+	SimpleSet skip = findSuperinterfaceCollisions(superclass, superInterfaces);
+	for (int i = 0; i < nextPosition; i++) {
+		superType = interfacesToVisit[i];
+		if (superType.isValidBinding()) {
+			if (skip != null && skip.includes(superType)) continue;
+			if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES) {
+				int itsLength = itsInterfaces.length;
+				if (nextPosition + itsLength >= interfacesToVisit.length)
+					System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
+				nextInterface : for (int a = 0; a < itsLength; a++) {
+					ReferenceBinding next = itsInterfaces[a];
+					for (int b = 0; b < nextPosition; b++)
+						if (next == interfacesToVisit[b]) continue nextInterface;
+					interfacesToVisit[nextPosition++] = next;
 				}
 			}
-		}
-	}
 
-	// bit reinitialization
-	for (int i = 0; i <= lastPosition; i++) {
-		ReferenceBinding[] interfaces = interfacesToVisit[i];
-		for (int j = 0, length = interfaces.length; j < length; j++)
-			interfaces[j].tagBits &= ~InterfaceVisited;
+			MethodBinding[] methods = superType.unResolvedMethods();
+			nextMethod : for (int m = methods.length; --m >= 0;) { // Interface methods are all abstract public
+				MethodBinding inheritedMethod = methods[m];
+				MethodBinding[] existingMethods = (MethodBinding[]) this.inheritedMethods.get(inheritedMethod.selector);
+				if (existingMethods == null) {
+					existingMethods = new MethodBinding[] {inheritedMethod};
+				} else {
+					int length = existingMethods.length;
+					// look to see if any of the existingMethods implement this inheritedMethod
+					for (int e = 0; e < length; e++)
+						if (isInterfaceMethodImplemented(inheritedMethod, existingMethods[e], superType))
+							continue nextMethod; // skip interface method with the same signature if visible to its declaringClass
+					System.arraycopy(existingMethods, 0, existingMethods = new MethodBinding[length + 1], 0, length);
+					existingMethods[length] = inheritedMethod;
+				}
+				this.inheritedMethods.put(inheritedMethod.selector, existingMethods);
+			}
+		}
 	}
 }
 void computeMethods() {
@@ -565,13 +618,20 @@ ProblemReporter problemReporter(MethodBinding currentMethod) {
 		reporter.referenceContext = currentMethod.sourceMethod();
 	return reporter;
 }
+SimpleSet findSuperinterfaceCollisions(ReferenceBinding superclass, ReferenceBinding[] superInterfaces) {
+	return null; // noop in 1.4
+}
+boolean reportIncompatibleReturnTypeError(MethodBinding currentMethod, MethodBinding inheritedMethod) {
+	problemReporter(currentMethod).incompatibleReturnType(currentMethod, inheritedMethod);
+	return true;
+}
 ReferenceBinding[] resolvedExceptionTypesFor(MethodBinding method) {
 	ReferenceBinding[] exceptions = method.thrownExceptions;
-	if ((method.modifiers & CompilerModifiers.AccUnresolved) == 0)
+	if ((method.modifiers & ExtraCompilerModifiers.AccUnresolved) == 0)
 		return exceptions;
 
 	if (!(method.declaringClass instanceof BinaryTypeBinding))
-		return TypeConstants.NoExceptions; // safety check
+		return Binding.NO_EXCEPTIONS; // safety check
 
 	for (int i = exceptions.length; --i >= 0;)
 		exceptions[i] = BinaryTypeBinding.resolveType(exceptions[i], this.environment, true);
